from enum import Enum


class DataType(Enum):
    """Represents all existing data types used in i/o streams of the
    model."""

    INT4 = "int4"
    INT8 = "int8"
    INT32 = "int32"
    INT64 = "int64"
    UINT8 = "uint8"
    UINT16 = "uint16"
    UINT32 = "uint32"
    UINT64 = "uint64"
    FLOAT32 = "float32"
    FLOAT16 = "float16"
    FLOAT64 = "float64"
    BOOLEAN = "boolean"
    STRING = "string"
